import functools
import json
import urllib.parse

from dataclasses import dataclass
from datetime import datetime
from typing import (
    Any,
    Dict,
    Iterator,
    List,
    Mapping,
    Tuple,
    Union,
)
from urllib.error import URLError

from tofh.plugins.mq.event import Event as TofhEvent

from .error import (
    CorruptPayload,
    InvalidTimestamp,
    InvalidURL,
    MissingKeys,
)


__all__ = [
    "Event",
    "tofh_to_cerebrum_converter",
]


EventName = str
Attributes = List[str]
Objects = List[Any]


@dataclass
class Event:
    """Represents a Cerebrum SCIM notification event received from MQ."""

    id: str
    event: str
    domain: str
    subject: str
    attributes: List[str]
    objects: List[Any]
    issuer: str
    audience: str
    exchange: str
    routing_key: str
    timestamp: datetime

    def __repr__(self):
        return (
            f"<{self.__class__.__name__}"
            f"id={self.id}, "
            f"event={self.event}, "
            f"routing_key={self.routing_key}>"
        )

    @classmethod
    def from_tofh_event(cls, event: TofhEvent):
        try:
            body = json.loads(event.body)
            sub = body.pop("sub")
        except json.JSONDecodeError as exc:
            raise CorruptPayload(event.body) from exc
        except KeyError as exc:
            raise MissingKeys(exc.args) from exc

        try:
            url = urllib.parse.urlparse(sub)
            parts = url.path.split("/")
            after_base_marker = parts.index("v1")
            _resource_type, subject = parts[after_base_marker + 1 :]
        except (AssertionError, AttributeError, URLError, ValueError) as exc:
            raise InvalidURL(sub) from exc

        try:
            # SCIM concession: we consider only the first event
            event_name, (attrs, objs) = next(iter_events(body))
            return cls(
                id=body.pop("jti"),
                event=event_name,
                domain=body.pop("resourceType"),
                subject=subject,
                attributes=attrs,
                objects=objs,
                issuer=body.pop("iss"),
                audience=body.pop("aud"),
                exchange=event.exchange,
                routing_key=event.routing_key,
                timestamp=datetime.fromtimestamp(body.pop("iat")),
            )
        except KeyError as exc:
            raise MissingKeys(exc.args) from exc


def iter_events(
    body: Dict[str, Any]
) -> Iterator[Tuple[EventName, Tuple[Attributes, Objects]]]:
    """
    Extracts attributes and objects associated with events from the
    Cerebrum notification.

    Event URIs are references to data attributes or objects stored
    at the top-level.  `body` is mutated by removing the `eventUris`
    field along with each of the associated SCIM data entries
    prefixed `urn:ietf:params:event:SCIM:*`.

    :raises KeyError:
        If `eventUris` or any of the SCIM data keys it lists are missing.
    :raises TooManyEvents:
        If the notification does not hold exactly one event.
    """
    for event_uri in body.pop("eventUris"):
        event_type = event_uri.split(":")[-1]
        data = body.get(event_uri, {})
        attrs = data.get("attributes", [])
        objs = data.get("objects", [])
        yield event_type, (attrs, objs)


def tofh_to_cerebrum_converter(func):
    """Decorator function for converting ``tofh.plugins.mq.event.Event`` into ``Event``."""

    @functools.wraps(func)
    def convert(tofh_event: TofhEvent):
        return func(Event.from_tofh_event(tofh_event))

    return convert
