"""Entry-points for zoom-deprovisioner-ms daemons."""

import argparse
import os
import pprint
import sys

import tofh.config

from tofh.plugins import mq

import dezoom
import dezoom.report

from dezoom import pid


default_pidfile = os.path.join(os.getcwd(), "dezoom-router.pid")


class UsageExit(SystemExit):
    code = 64

    def __init__(self, message):
        eprint(f"dezoom: error: {message}")


# TODO(andretol): this does not print the Celery or tofh configurations
# because they are not yet defined in dezoom.Config
def config(opts, argv):
    """Pretty-prints the current config."""
    cfg = dezoom.config.get(opts.config)
    dezoom.config.display(cfg)


def router(opts, argv):
    """
    Launches an MQ consumer (tofh.plugins.mq) for routing incoming messages
    to their correct destination defined by tofh_config.
    """
    with pid.Lock(opts.pidfile):
        config = tofh.config.get_config(opts.config)
        consumer = mq.get_consumer(config)
        try:
            consumer.run()
        except KeyboardInterrupt:
            consumer.stop()


def worker(opts, celery_args):
    """Launches a Celery worker for processing MQ messages."""
    argv = ["celery", "worker", "--app", "dezoom.tofh"] + celery_args
    env = os.environ.copy()
    env["TOFH_CONFIG"] = opts.config
    os.execvpe("celery", argv, env)


def report(opts, argv):
    """Trigger reporting."""
    filename = opts.filename
    cfg = dezoom.config.get(opts.config)
    ctx = dezoom.Context(config=cfg, destructive=False)
    dezoom.report.check_all_users(ctx=ctx, filename=filename)


def inspect_user(opts, argv):
    """Trigger inspection of a single user."""
    cfg = dezoom.config.get(opts.config)
    username = opts.username
    ctx = dezoom.Context(config=cfg, destructive=False)

    from cerebrum_client.models import Account

    account = ctx.cerebrum.get_account(username)
    assert isinstance(account, Account)

    from dezoom.lifecycle import inspect

    inspect_result = inspect(ctx, account)
    print("inspect result", inspect_result)


def process_user(opts, argv):
    """Trigger processing of a single user."""
    cfg = dezoom.config.get(opts.config)
    username = opts.username
    destructive = opts.destructive
    ctx = dezoom.Context(config=cfg, destructive=destructive)

    from cerebrum_client.models import Account

    account = ctx.cerebrum.get_account(username)
    assert isinstance(account, Account)

    from dezoom.lifecycle import start

    result = start(ctx, account, allow_deletion=destructive)
    print("result", result)


def batch_deprovision(opts, argv):
    """Trigger batch deprovisioning."""
    cfg = dezoom.config.get(opts.config)
    destructive = opts.destructive
    ctx = dezoom.Context(config=cfg, destructive=destructive)

    from dezoom.consumer import deprovision_batch

    deprovision_batch(ctx=ctx)


def batch_process_all(opts, argv):
    """Trigger batch processing of all Zoom users."""
    cfg = dezoom.config.get(opts.config)
    ctx = dezoom.Context(config=cfg, destructive=False)

    from dezoom.consumer import process_all_batch

    process_all_batch(ctx=ctx)


targets = {
    "config": config,
    "report": report,
    "router": router,
    "inspect": inspect_user,
    "process": process_user,
    "worker": worker,
    "batch-deprovision": batch_deprovision,
    "batch-process-all": batch_process_all,
}


def main(inargs=None):
    import logging

    logging.basicConfig(level=logging.DEBUG)

    parser = argparse.ArgumentParser(description="Deprovisioner of Zoom accounts.")
    parser.add_argument(
        "--config",
        "-c",
        metavar="FILE",
        required=True,
        help="configuration for dezoom, tofh, and Celery",
    )
    parser.add_argument("--version", action="version", version=dezoom.__version__)

    subparsers = parser.add_subparsers(dest="target")

    config_parser = subparsers.add_parser("config", help="display configuration")

    router_parser = subparsers.add_parser("router", help="start a tofh router process")
    router_parser.add_argument(
        "--pidfile",
        metavar="FILE",
        default=default_pidfile,
        help="global lock file to prevent multiple router processes (default: %(default)s)",
    )

    worker_parser = subparsers.add_parser(
        "worker", help="start a Celery worker process"
    )

    report_parser = subparsers.add_parser(
        "report", help="make reports about the current state in Zoom"
    )
    report_parser.add_argument(
        "--filename", type=str, required=True, help="write report to this path",
    )

    inspect_parser = subparsers.add_parser("inspect", help="inspect a single username")
    inspect_parser.add_argument(
        "--username", type=str, required=True, help="username to inspect",
    )

    process_parser = subparsers.add_parser(
        "process", help="process a single username, running lifecycle logic"
    )
    process_parser.add_argument(
        "--username", type=str, required=True, help="username to inspect",
    )
    process_parser.add_argument(
        "--destructive",
        action="store_true",
        help="allow destructive API calls to Zoom",
    )

    batch_deprovision_parser = subparsers.add_parser(
        "batch-deprovision", help="trigger batch deprovisioning"
    )
    batch_deprovision_parser.add_argument(
        "--destructive",
        action="store_true",
        help="allow destructive API calls to Zoom",
    )

    batch_process_all_parser = subparsers.add_parser(
        "batch-process-all", help="trigger processing of all zoom users"
    )

    opts, argv = parser.parse_known_args(inargs)
    try:
        targets[opts.target](opts, argv)
    except pid.AlreadyRunning:
        raise UsageExit(f"{opts.target} already running")


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


if __name__ == "__main__":
    main(sys.argv[1:])
