"""
Configuration is written in YAML and holds information relevant to
the operation of zoom-deprovisioner-ms.  This could be authentication
tokens for accessing Cerebrum's and Zoom's client HTTP APIs, along
with configuration for tofh and Celery, since they share the same
format.

Configuration can be loaded programmatically::

	import dezoom.config
	config = dezoom.config.get(filename)

The `filename` argument is optional, and if omitted the configuration
will be looked up here:

    1. The `DEZOOM_CONFIG` environment variable.
    2. A file named *dezoom.yml* in the first matching location:
        1. the current working directory
        2. `$HOME/dezoom/dezoom.yml`
        3. `/etc/dezoom/dezoom.yml` if running outside a venv,
           or the same path relative to the root of the venv.

You may also compute the lookup paths for diagnostic purposes this
way::

	import dezoom.config
	for path in dezoom.config.iter():
	    print(path)

An example configuration file might look like this::

	dezoom:
	  enabled: true
	  license_downgrade_delay: 7
	  account_deletion_delay: 30
	  seen_group_id: foo
	  downgraded_group_id: bar
	  deprovision_group_id: baz
	  realm: uio.no

	cerebrum:
	  url: https://gw-uio.intark.uh-it.no/cerebrum/
	  headers:
	    x-gravitee-api-key: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx

	zoom:
	  url: https://gw-uio.intark.uh-it.no/zoom/
	  headers:
	    x-gravitee-api-key: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
	  transport: dezoom.zoom.transport.NoopHTTPProtocol
"""

import functools
import importlib
import json
import os
import pprint
import sys
import yaml

from urllib.parse import urlparse
from typing import (
    Any,
    Dict,
    Iterable,
    Optional,
    TextIO,
    Tuple,
    Type,
)

from pydantic import (
    BaseModel,
    root_validator,
    validator,
)

from dezoom.zoom.transport import HTTPWireProtocol
from dezoom.zoom.typing import WireProtocol
from dezoom.zoom.v2.users import LicenseType


__all__ = [
    "Config",
    "display",
    "get",
    "iter",
]


CONFIG_FILE = "dezoom.yml"
CONFIG_ENV = "DEZOOM_CONFIG"
FALLBACK_LOOKUP_ORDER = [
    os.getcwd(),
    os.path.expanduser("~/dezoom"),
    os.path.join(
        os.path.join("" if sys.prefix == "/usr" else sys.prefix, "etc"), "dezoom"
    ),
]


class ConfigBaseModel(BaseModel):
    def json(self):
        return json.loads(super().json())

    class Config:
        json_encoders = {type: lambda cls: f"{cls.__module__}.{cls.__qualname__}"}


class CerebrumConfig(ConfigBaseModel):
    url: str
    headers: Dict[str, str] = {}


class DezoomConfig(ConfigBaseModel):
    enabled: bool = False
    license_downgrade_delay: int = 7
    account_deletion_delay: int = 30
    full_license_type: LicenseType = LicenseType.ON_PREM
    seen_group_id: str
    downgraded_group_id: str
    deprovision_group_id: str
    realm: str
    realm_short: str

    @root_validator
    def check_delays(cls, values):
        license_downgrade_delay = values.get("license_downgrade_delay")
        account_deletion_delay = values.get("account_deletion_delay")
        if license_downgrade_delay > account_deletion_delay:
            raise ValueError(
                "account_deletion_delay must be higher than license_downgrade_delay"
            )
        return values

    @validator("full_license_type", pre=True)
    def license_type_to_enum_instance(cls, v: Any):
        if isinstance(v, LicenseType):
            return v
        if isinstance(v, int):
            return LicenseType(v)
        raise TypeError("expected a LicenseType instance or numeric identifier")


class ZoomConfig(ConfigBaseModel):
    url: str
    headers: Dict[str, str] = {}
    transport: Optional[Type[WireProtocol]] = HTTPWireProtocol

    def __repr__(self):
        # transport is a class, not an instance, and the correct representation
        # in this edge case is a str, e.g. "dezoom.zoom.transport.HTTPWireProtocol"
        return (
            f"{self.__module__}.{type(self).__name__}("
            f"url={self.url!r}, "
            f"headers={self.headers!r}, "
            f"transport='{self.transport.__module__}.{self.transport.__name__}'"
            ")"
        )

    @validator("url")
    def valid_url(cls, v):
        assert valid_url(v), f"Invalid URL: {v!r}"
        return v

    @validator("transport", pre=True)
    def transport_into_class(cls, v: Any):
        if not isinstance(v, str):
            if issubclass(v, WireProtocol):
                return v
            raise TypeError("expected implementation of WireProtocol")
        modname, _, clsname = v.rpartition(".")
        mod = importlib.import_module(modname)
        return getattr(mod, clsname)


class Config(ConfigBaseModel):
    cerebrum: CerebrumConfig
    dezoom: DezoomConfig
    zoom: ZoomConfig

    @classmethod
    def from_yaml(cls, s: str) -> "Config":
        return cls(**yaml.load(s, Loader=yaml.FullLoader))

    @classmethod
    def from_file(cls, filename: str) -> "Config":
        with open(filename) as f:
            return cls.from_yaml(f.read())


def iter(lookup_order: Optional[Tuple[str]] = None) -> Iterable[str]:
    if CONFIG_ENV in os.environ:
        yield os.path.abspath(os.environ[CONFIG_ENV])
    for path in lookup_order or FALLBACK_LOOKUP_ORDER:
        yield os.path.abspath(os.path.join(path, CONFIG_FILE))


@functools.lru_cache()
def get(filename: str = None, lookup_order: Optional[Tuple[str]] = None,) -> Config:
    if filename is None:
        for filename in filter(os.path.exists, iter(lookup_order)):
            break
    if filename:
        return Config.from_file(filename)
    raise IOError(
        "Unable to locate configuration file dezoom.yml "
        f"(search path: {', '.join(list(iter(lookup_order)))})"
    )


def display(config: Config, file: TextIO = sys.stdout):
    """Pretty-prints the configuration."""
    pprint.pprint(config.dict(), stream=file)


def valid_url(url: str) -> bool:
    scheme, netloc, path, _, _, _ = urlparse(url)
    return all([scheme, netloc, path])


# python -m dezoom.config
def main(inargs=None):
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "filename", nargs="?", help="parse YAML file and pretty-print result",
    )
    args = parser.parse_args(inargs)

    if args.filename:
        cfg = get(args.filename)
        display(cfg)
    else:
        for f in iter():
            missing = " (missing)" if not os.path.exists(f) else ""
            print(f + missing)


if __name__ == "__main__":
    main(sys.argv[1:])
