"""
Consumers are either reactive or scheduled tasks that control Zoom
user deprovisioning.

The deprovisioning process is controlled by three distinct tasks.
The first and most important, is the lifecycle_batch() which steers
the account through the different stages in the process.  The second
is account_modify() which puts accounts in the batch of accounts
for lifecycle_batch() to inspect.  The third and final task is
deprovision_batch() which simply will delete Zoom users that the
lifecycle_batch() has previously decided are ready for deprovisioning.

Setting a Cerebrum account's expiry date will schedule its associated
Zoom user for deprovisioning.  We refer to accounts that have an
expiry date set but which does not yet meet the license_downgrade_delay
threshold, as "seen" accounts.  These will be revisited regularly
to track their progress.

Accounts that do meet the configured license_downgrade_delay threshold
are downgraded to a basic license, but not deleted immediately.
This users a grace period so that their scheduled meetings are not
removed from their calendars.  We refer to these users as "downgraded"
users.

Finally, accounts that meet the account_deletion_delay are scheduled
for deprovisioning by lifecycle_batch().  This means they are moved
to a deprovisioning queue, whcih deprovision_batch() will act upon
at a less frequent interval.

If at any point during this process a Cerebrum account regains its
active status by losing its expiry date, it will also be upgraded
to a full license and removed from the deprovisioning process
altogether.
"""

import logging

from itertools import chain

from result.result import Err, Ok

from dezoom import (
    AccountNotFound,
    Context,
    task,
)
from dezoom.lifecycle import (
    AUTOGENERATED_EMAIL,
    delete_zoom_user,
    remove_zoom_user_realm,
    parse_email,
    start,
)
from dezoom.logging import ScopedLogger
from dezoom.zoom.v2 import NotFound


logger = logging.getLogger(__name__)


@task
def account_modify(ctx: Context, uname: str, *args, **kwargs):
    """
    Reacts to Cerebrum account expiry changes and schedules their
    associated Zoom users for deprovisioning.

    This task reacts to *.cerebrum.scim.accounts.modify notifications
    from Cerebrum, looking for accounts which have had an expiry
    date set.  This indicates that the account is scheduled for
    deactivation at some future point in time, and is the cue for
    us to begin deprovisioning its associated Zoom teleconferencing
    user account.
    """
    log = ScopedLogger(logger, uname)

    attributes = kwargs.get("attributes", [])
    if "expireDate" not in attributes:
        return log.info("expiry date is unchanged")

    account = ctx.cerebrum.get_account(uname)
    if not account:
        raise AccountNotFound(uname)

    start(ctx, account, allow_deletion=False)


# @task
def lifecycle_batch(ctx: Context, *args, **kwargs):
    """
    Tracks deprovisioning progress of Zoom users.

    We keep tabs on the associated Zoom users of Cerebrum accounts
    that are soon to expire or expired.  Since deprovisioning is
    graded, we delay downgrading their licenses until the
    license_downgrade_delay is met.  Correspondingly, we delay
    scheduling users for deletion until they meet the
    account_deletion_delay requirement.

    By inspecting users in the seen and downgraded queues, we can
    follow up on their progress.  As their associated Cerebrum
    account's expiry date moves meets the different thresholds, the
    users are either downgraded (if they meet the license_downgrade_delay),
    upgraded (if their expiry date has been removed), or scheduled
    for deprovisioning (if the account_deletion_delay is met).

    This batch job is non-destructive, meaning accounts that meet
    the account_deletion_delay are not deleted but instead scheduled
    for later removal by the deprovision_batch() task.  That task
    runs less often that lifecycle_batch(), achieving the effect
    that Zoom users are not deleted immediately when we receive an
    account modification notification from Cerebrum with an expiry
    date that meets the requirements.
    """
    all_candidates = list(
        chain(
            ctx.zoom.groups.list_members(ctx.config.dezoom.seen_group_id),
            ctx.zoom.groups.list_members(ctx.config.dezoom.downgraded_group_id),
        )
    )
    for user in all_candidates:
        parsed = parse_email(ctx, user.email)
        if isinstance(parsed, Err):
            logger.warning(f"{user.id}: {parsed.value} ({user.email})")
            continue
        assert isinstance(parsed, Ok)
        uname = parsed.value
        account = ctx.cerebrum.get_account(uname)
        if not account:
            raise AccountNotFound(uname)
        # TODO(andretol): fan out a tofh task for each user
        start(ctx, account, allow_deletion=False)


# @task
def deprovision_batch(ctx: Context, *args, **kwargs):
    """
    Deletes Zoom users scheduled for deprovisioning.

    By inspecting the Zoom users in the deprovisioning group, it
    is decided whether if the Zoom user still meets the conditions
    for deprovisioning.  If it does, the user account is deleted.

    If however the associated Cerebrum account no longer has an
    expiry date, or its expiry date does not meet the
    account_deletion_delay threshold, it may be upgraded and restored
    to a full license.
    """
    if ctx.config.dezoom.enabled:
        logger.info("destructive deprovisioning")
    else:
        logger.info("dry run deprovisioning")
    all_candidates = list(
        ctx.zoom.groups.list_members(ctx.config.dezoom.deprovision_group_id)
    )
    for user in all_candidates:
        logger.info(f"processing {user.id} {user.email}")
        parsed = parse_email(ctx, user.email)
        if isinstance(parsed, Err):
            err = parsed.value
            if err == AUTOGENERATED_EMAIL:
                delete_zoom_user(ctx, user)
                continue
            else:
                logger.warning(f"{user.id}: {err} ({user.email})")
                continue
        assert isinstance(parsed, Ok)
        uname = parsed.value
        account = ctx.cerebrum.get_account(uname)
        if not account:
            raise AccountNotFound(uname)
        # TODO(andretol): fan out a tofh task for each user
        start(ctx, account, allow_deletion=ctx.config.dezoom.enabled)


def process_all_batch(ctx: Context, *args, **kwargs):
    """
    Lists all users in Zoom and fans out an internal processing task for each user.
    """
    assert ctx.destructive is False

    log = ScopedLogger(logger, "")

    log.info("fetching all users")
    users = []
    count = 0
    for user in ctx.zoom.users.list():
        users.append(user)
        count += 1
        if count % 1000 == 0:
            log.info("... %s", count)
    log.info("fetched", count, "users")

    from dezoom.tofh import app as celery_app

    task = celery_app.tasks["internal-process-user"]
    options = {}
    if task.priority:
        options["priority"] = task.priority

    for user in users:
        log.info(
            "queueing internal task for user with id=%r email=%r", user.id, user.email
        )
        task.apply_async(args=None, kwargs={"user_id": user.id}, **options)


@task
def internal_process_user(ctx: Context, user_id: str, *args, **kwargs):
    """
    Reacts to an internally queued processing task.
    """
    log = ScopedLogger(logger, user_id)
    log.info(f"processing {user_id}")

    try:
        user = ctx.zoom.users.get(user_id)
    except NotFound:
        log.info("no associated Zoom user")
        return

    parsed = parse_email(ctx, user.email)
    if isinstance(parsed, Err):
        err = parsed.value
        if err == AUTOGENERATED_EMAIL:
            # TODO: move to deprovisioning group?
            log.warning(f"{err} ({user.email})")
        else:
            log.warning(f"{err} ({user.email})")
        return
    assert isinstance(parsed, Ok)
    uname = parsed.value
    account = ctx.cerebrum.get_account(uname)
    if not account:
        raise AccountNotFound(uname)

    start(ctx, account, allow_deletion=False)
