from functools import cached_property

from cerebrum_client.client import CerebrumClient

from dezoom import Config
from dezoom.zoom import ZoomClient


class Context:
    """
    Contexts carry state and other task-scoped values across Celery
    task boundaries.
    """

    def __init__(self, config: Config, destructive: bool = False):
        self.config: Config = config
        """Configuration this context is based upon."""

        self.destructive: bool = destructive
        """
        Whether context acts destructively, e.g. causes quarantined accounts
        to be either downgraded or deprovisioned if they meet the license downgrade
        or account deletion delays.
        """

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, Context)
            and self.config == other.config
            and self.destructive == other.destructive
        )

    def __str__(self) -> str:
        return (
            f"<{self.__class__.__name__} "
            f"{'destructive' if self.destructive else 'non-destructive'} "
            f"object at {hex(id(self))}>"
        )

    def __repr__(self) -> str:
        return "%s(%r, destructive=%r)" % (
            self.__class__.__name__,
            self.config,
            self.destructive,
        )

    @cached_property
    def cerebrum(self) -> CerebrumClient:
        """Create or return the current Cerebrum HTTP API session."""
        return CerebrumClient(**self.config.cerebrum.dict())

    @cached_property
    def zoom(self) -> ZoomClient:
        """Create or return the current Zoom API session."""
        return ZoomClient(**self.config.zoom.dict())

    @classmethod
    def xerox(cls, other: "Context") -> "Context":
        """Makes a copy of ``other``."""
        return cls(other.config, destructive=other.destructive)
