"""The lifecycle contains the core business logic for Zoom user deprovisioning."""

from dezoom.zoom.v2.groups import GroupMember
from dezoom.zoom.v2.users import User
import logging

from datetime import (
    date,
    datetime,
    timedelta,
)
from typing import (
    Any,
    Iterable,
    Union,
)
from result import Ok, Err, Result

from cerebrum_client.models import Account

from dezoom import Context
from dezoom.logging import ScopedLogger
from dezoom.zoom.v2 import (
    GroupMemberNotFound,
    LicenseType,
    NotFound,
)


logger = logging.getLogger(__name__)


# TODO(andretol): how does one define a function prototype
# returning its own prototype, e.g. type stateFn(...) -> stateFn?
StateFn = Any

AUTOGENERATED_EMAIL = "autogenerated email"
INVALID_EMAIL = "invalid email"
UNEXPECTED_REALM = "unexpected realm"


def make_zoom_user_id(ctx: Context, name: str) -> str:
    """Adds realm to a Cerebrum account name so that it matches the Zoom user IDs."""
    return f"{name}@{ctx.config.dezoom.realm}"


def remove_zoom_user_realm(ctx: Context, id: str) -> str:
    """Removes realm from a Zoom user ID so that it matches the Cerebrum account name."""
    return id.removesuffix(f"@{ctx.config.dezoom.realm}")


def parse_email(ctx: Context, email: str) -> Result[str, str]:
    """
    Attempts to extract the Cerebrum account name from an email address,
    which is expected to be on the form 'uid@example.com',
    where 'example.com' is the configured realm (ctx.config.dezoom.realm).

    Sometimes, Zoom creates a user without a valid email address,
    which will be on the form 'example__ababababababababababababababababababababab',
    where 'example' is the Zoom instance ID (ctx.config.dezoom.realm_short).
    """
    wanted_realm = ctx.config.dezoom.realm
    missing_email_prefix = f"{ctx.config.dezoom.realm_short}__"
    if email.startswith(missing_email_prefix) and "@" not in email:
        return Err(AUTOGENERATED_EMAIL)
    if "@" not in email:
        return Err(INVALID_EMAIL)
    local, realm = email.split("@")
    if realm != wanted_realm:
        return Err(UNEXPECTED_REALM)
    return Ok(local)


def start(
    ctx: Context, account: Account, allow_deletion: bool = False
) -> Iterable[StateFn]:
    """
    Start the state machine and run until completion.

    Setting allow_deletion will run the lifecycle to run in destructive
    mode and cause the account to be deprovisioned if it meets the
    acction_deletion_delay threshold.  In non-destructive mode,
    which is the default, an account that meets the same threshold
    will be put in the deprovisioning group for later removal.

    Returns the chain of states that were executed.
    """
    # avoid mutating passed ctx
    ctx = Context.xerox(ctx)
    ctx.destructive = allow_deletion

    elapsed_states = []
    state = inspect
    while state != done:
        try:
            elapsed_states.append(state)
            state = state(ctx, account)
        except Exception:
            logger.error(f"{state}: unexpected error", exc_info=True)
            raise

    return tuple(elapsed_states)


def inspect(ctx: Context, account: Account) -> StateFn:
    """Inspect a Cerebrum account and determine its associated Zoom user's fate."""
    downgrade_delay = ctx.config.dezoom.license_downgrade_delay
    deletion_delay = ctx.config.dezoom.account_deletion_delay

    # active accounts should have a full license
    if account.active and account.expire_date is None:
        return upgrade

    assert isinstance(account.expire_date, datetime)
    expire_date = account.expire_date.date()
    today = date.today()

    # if the number of days in deletion_delay has passed we want to deprovision the user
    # by either deleting it (if destructive) or adding it to the deprovisioning queue
    deletion_cutoff = expire_date + timedelta(days=deletion_delay)
    if today > deletion_cutoff:
        return deprovision

    # if the number of days in downgrade_delay has passed,
    # we want to downgrade the user to a basic license
    downgrade_cutoff = expire_date + timedelta(days=downgrade_delay)
    if today > downgrade_cutoff:
        return downgrade

    # account has an expire date, but it is not covered
    # by the above inclusion clauses for deprovisioning
    return seen


def upgrade(ctx: Context, account: Account) -> StateFn:
    """
    Ensures account has a full license.

    Does nothing if the Cerebrum account does not have
    an associated Zoom user, or if it already has a full license.
    """
    user_id = make_zoom_user_id(ctx, account.name)
    log = ScopedLogger(logger, user_id)
    try:
        user = ctx.zoom.users.get(user_id)
    except NotFound:
        log.info("no associated Zoom user")
        return done
    if user.type == LicenseType.BASIC:
        wanted_license_type = ctx.config.dezoom.full_license_type
        log.info(f"upgrading to {wanted_license_type} license")
        ctx.zoom.users.set_license_type(user.id, wanted_license_type)
    else:
        log.info("already licensed")
    return move(None)


def downgrade(ctx: Context, account: Account) -> StateFn:
    """
    Ensures account has a basic license.

    Does nothing if the Cerebrum account does not have
    an associated Zoom user, or if it already has a basic license.
    """
    user_id = make_zoom_user_id(ctx, account.name)
    log = ScopedLogger(logger, user_id)
    try:
        user = ctx.zoom.users.get(user_id)
    except NotFound:
        log.info("no associated Zoom user")
        return done
    if user.type != LicenseType.BASIC:
        log.info(f"downgrading to {LicenseType.BASIC} license")
        ctx.zoom.users.set_license_type(user.id, LicenseType.BASIC)
    else:
        log.info("already unlicensed")
    return move(ctx.config.dezoom.downgraded_group_id)


def seen(ctx: Context, account: Account) -> StateFn:
    """
    Schedules Zoom user for deprovisioning.

    This happens when the account does not yet meet
    license_downgrade_delay or account_deletion_delay,
    but has an expiry date set.

    Scheduling entails removing user from all other queues.
    """
    logger.info(f"{account.name}: enqueuing for downgrade")
    return move(ctx.config.dezoom.seen_group_id)


def deprovision(ctx: Context, account: Account) -> StateFn:
    """
    Deprovisions Zoom user if operating within a destructive context,
    otherwise adds it to the deprovisioning queue.

    This happens when the account has met the account_deletion_delay.
    """
    user_id = make_zoom_user_id(ctx, account.name)
    log = ScopedLogger(logger, user_id)
    if ctx.destructive:
        try:
            user = ctx.zoom.users.get(user_id)
        except NotFound:
            log.info("no associated Zoom user")
            return done
        log.info("deleting Zoom user")
        ctx.zoom.users.delete(user.id)
        return done
    else:
        log.info(
            "deletion threshold met, but skipping deprovisioning "
            "because lifecycle is operating in non-destructive mode",
        )
        return move(ctx.config.dezoom.deprovision_group_id)


def delete_zoom_user(ctx: Context, user: Union[User, GroupMember]) -> bool:
    """
    Special method for deleting Zoom users directly. Used when we don't have a
    Cerebrum account to feed into the lifecycle methods, but want to delete
    the user anyway.
    """
    log = ScopedLogger(logger, user.id)
    log.info(f"bypassing lifecycle, want to delete the user identified as {user}")
    if ctx.destructive:
        log.info("deleting Zoom user")
        ctx.zoom.users.delete(user.id)
        return True
    else:
        log.info("skipping deletion because we're operating in non-destructive mode",)
        return False


class move:
    """
    Moves associated Zoom user to target queue.

    Although a user, under normal circumstances, would not be member
    of more than maximum one group, this move operation is designed
    to balance the queues in the off chance of this being the case.
    This means that if the user is removed all other queues than
    the target queue.

    If no target is given as None, the user is removed from all queues.
    """

    def __init__(self, target_group: Union[str, None]):
        self.target = target_group

    def __eq__(self, other):
        return isinstance(other, move) and self.target == other.target

    def __repr__(self):
        return f"{type(self).__name__}({self.target!r})"

    def __call__(self, ctx: Context, account: Account) -> StateFn:
        all_groups = {
            ctx.config.dezoom.seen_group_id,
            ctx.config.dezoom.downgraded_group_id,
            ctx.config.dezoom.deprovision_group_id,
        }
        targets = set([self.target])

        user_id = make_zoom_user_id(ctx, account.name)
        log = ScopedLogger(logger, user_id)
        try:
            user = ctx.zoom.users.get(user_id)
        except NotFound:
            log.info("no associated Zoom user")
            return done

        user_groups = set(user.group_ids or []).intersection(all_groups)

        for group_id in user_groups - targets:
            log.info(f"removing from group {group_id}")
            try:
                ctx.zoom.groups.remove_member(group_id, user.id)
            except GroupMemberNotFound:
                log.warning(f"not a member of group {group_id}")

        if self.target is not None:
            if self.target in user_groups:
                log.info(f"already member of {self.target}")
            else:
                log.info(f"adding to group {self.target}")
                ctx.zoom.groups.add_member(self.target, user.id)

        return done


def done(ctx: Context, account: Account) -> StateFn:
    """Indicates the state machine is complete."""
    raise AssertionError("unreachable")
