"""
Provides a zero-dependency PID lock file implementation.

PID files contain a program's process identifier and is an
out-of-process technique allowing long running processes to be
self-aware.  A server process may inspect it to stop itself, or
have a heuristic that its other instance is already running.  They
may also be used together with pkill(1):

	% pkill -F <pidfile>

``pid.Lock`` provides a context manager for preventing accidentally
starting the same process twice, when this is unwanted behaviour.

Starting the following program twice within 30 seconds, will result
in the second process exiting with an ``AlreadyRunningError``
exception::

    import pid
    import sys
    import time

    with pid.Lock(sys.argv[0]):
        time.sleep(30)
"""

import errno
import os


class AlreadyRunning(Exception):
    """Raised when a process is already running and protected by a PID file."""


class Lock:
    def __init__(self, fname):
        self.file = fname

    @property
    def pid(self):
        if not os.path.exists(self.file):
            return None
        try:
            with open(self.file) as f:
                return int(f.read())
        except (OSError, ValueError):
            return None

    @property
    def is_running(self):
        return os.path.exists(self.file)

    def __enter__(self):
        if self.is_running:
            raise AlreadyRunning
        with open(self.file, "w") as f:
            f.write(str(os.getpid()))
        return self

    def __exit__(self, *args):
        if os.path.exists(self.file):
            try:
                os.remove(self.file)
            except OSError as e:
                if e.errno != errno.ENOENT:
                    raise


def exists(pid):
    """Check if PID exists in the current process table."""
    try:
        os.kill(pid, 0)
    except ProcessLookupError:
        return False
    except PermissionError:
        return True
    # according to kill(2) possible errors are EINVAL, EPERM, ESRCH
    else:
        return True
