import functools

from types import SimpleNamespace
from typing import (
    Callable,
    Optional,
)

import dezoom
from dezoom import (
    Config,
    Context,
)


__all__ = ["task"]


def ensure_context(ns: SimpleNamespace, config: Optional[Config] = None):
    if not hasattr(ns, "context"):
        config = config or dezoom.config.get()
        ns.context = Context(config)
    return ns.context


def varardic_wraps(func):
    """Allows a decorator to be used both with and without arguments."""

    @functools.wraps(func)
    def call_or_wrap(*args, **kwargs):
        if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
            return func(args[0])
        else:
            return lambda real_func: func(real_func, *args, **kwargs)

    return call_or_wrap


@varardic_wraps
def task(func: Callable[[Context], None], config: Optional[Config] = None):
    """
    Decorator turning a Celery task into a context.

    Contexts are containers that carry state and other task-scoped
    values across task boundaries.  They act as stateful containers
    preserving different session data, such as configurations and
    API sessions.

    The decorator upgrades the task function's first argument from
    a ``celery.app.task.Task`` to a ``dezoom.Context``::

        import dezoom

        @dezoom.task
        def account_deactivate(ctx: dezoom.Context, uname: str, *args, **kwargs):
            pass

    Note that the decorator currently only works with stateful tasks.
    """

    @functools.wraps(func)
    def upgrade_context(task, *args, **kwargs):
        ctx = ensure_context(task, config=config)
        return func(ctx, *args, **kwargs)

    return upgrade_context
