from typing import (
    Dict,
    Optional,
    Type,
)

from . import v2
from .api import API
from .resource import Resource
from .transport import HTTPWireProtocol
from .typing import WireProtocol


# Mypy does not allow using modules as subtypes of protocols:
# https://github.com/python/mypy/issues/5018
supported_versions: Dict[str, API] = {"2": v2}  # type: ignore


class ZoomClient:
    """HTTP client for the Zoom video conferencing REST API."""

    api: API
    transport: WireProtocol

    _resource_cache: Dict[str, Resource] = {}

    def __init__(
        self,
        url: str,
        headers: Optional[Dict[str, str]] = None,
        version: str = "2",
        transport: Type[WireProtocol] = HTTPWireProtocol,
    ):
        """
        Builds a new Zoom API connection.

        The optional ``version`` parameter decides which version
        of the Zoom API to use.  Supported API versions are listed
        in ``dezoom.zoom.client.supported_versions``.

        :raises ValueError:
            If the provided ``url`` is not valid,
            or if the ``version`` is not supported.
        """
        try:
            self.api = supported_versions[version]
        except KeyError:
            raise ValueError(f"Unsupported API version: {version}")
        self.transport = transport(url, prefix=self.api.namespace, headers=headers)

    def __getattr__(self, name):
        try:
            return self._ensure_resource(name)
        except KeyError as err:
            raise AttributeError() from err

    def _ensure_resource(self, ns):
        if ns in self._resource_cache:
            return self._resource_cache[ns]
        cls = self.api.resources[ns]
        self._resource_cache[ns] = cls(self.transport)
        return self._ensure_resource(ns)

    def __str__(self):
        return f"<{self.__class__.__name__} {self.transport.url_for('/')}>"
