import json
import logging

from collections import defaultdict
from typing import (
    Any,
    Dict,
    Optional,
    cast,
)
from urllib.parse import urlparse

import requests

from dezoom.zoom.typing import (
    Method,
    Response,
    WireProtocol,
)
from dezoom.zoom.v2 import error


def slash_join(*parts: str) -> str:
    """
    Like urllib.parse.urljoin, except a leading slash in a part
    doesn't strip any previous paths.
    """
    return "/".join(part.strip("/") for part in parts)


class HTTPResponse(Response):
    def __init__(self, content: Optional[bytes] = b"", code: int = 200):
        self._code = code
        self._content = content

    def __str__(self):
        return f"<{type(self).__name__} {self.code} object at {hex(id(self))}>"

    @property
    def code(self) -> int:
        return self._code

    @property
    def content(self) -> Optional[bytes]:
        return self._content

    @property
    def text(self) -> str:
        if self.content is None:
            return str()
        return str(self.content, errors="replace")

    def json(self, **kwargs):
        return json.loads(self.text, **kwargs)

    def raise_if_error(self):
        if self.code not in (200, 201, 202, 204):
            raise error.from_response(self)

    @classmethod
    def from_json(cls, obj: Any, code: int = 200) -> Response:
        return cls.from_text(json.dumps(obj), code=code)

    @classmethod
    def from_text(cls, s: str, code: int = 200) -> Response:
        content = bytearray(s.encode())
        return cls(content, code)

    @classmethod
    def from_requests(cls, raw: requests.Response) -> Response:
        content, code = raw.content, raw.status_code
        return cls(content, code)

    @classmethod
    def empty(cls) -> Response:
        return cls()


class HTTPWireProtocol(WireProtocol):
    def __init__(self, url: str, prefix: str = "/", headers: Dict[str, str] = None):
        if not valid_url(url):
            raise ValueError(f"Invalid URL: {url!r}")
        self.url = slash_join(url, prefix)
        self.headers = defaultdict(str, headers or {})
        self.session = requests.Session()

    def __str__(self):
        return f"<{type(self).__name__} {self.url} object at {hex(id(self))}>"

    def __repr__(self):
        return (
            f"{self.__module__}.{type(self).__name__}("
            f"{self.url!r}, "
            f"headers={dict(self.headers)!r}"
            ")"
        )

    def url_for(self, path: str) -> str:
        return slash_join(self.url, path)

    def get(self, path: str, **query: Any) -> Response:
        return self.request(Method.GET, path, params=query)

    def patch(self, path: str, **fields: Any) -> Response:
        return self.request(Method.PATCH, path, json=fields)

    def post(self, path: str, **entity: Any) -> Response:
        return self.request(Method.POST, path, json=entity)

    def delete(self, path: str, **query: Any) -> Response:
        return self.request(Method.DELETE, path, params=query)

    def request(
        self,
        method: Method,
        path: str,
        params: Optional[Dict[str, str]] = None,
        json: Optional[Dict[str, Any]] = None,
    ) -> Response:
        raw = self._raw_request(method, path, params=params, json=json)
        resp = HTTPResponse.from_requests(raw)
        resp.raise_if_error()
        return resp

    def _raw_request(
        self,
        method: Method,
        path: str,
        params: Optional[Dict[str, str]] = None,
        json: Optional[Dict[str, Any]] = None,
    ) -> requests.Response:
        return self.session.request(
            method.value,
            self.url_for(path),
            headers=self.headers,
            params=params,
            json=json,
        )


class LoggingWireProtocol(WireProtocol):
    """Logs all HTTP requests and responses of ``delegate``."""

    def __init__(self, delegate: WireProtocol, logger: logging.Logger = None):
        self.delegate = delegate
        self.delegate_name = f"{delegate.__module__}.{type(delegate).__name__}"
        self.log = logger or logging.getLogger(__name__)

    def url_for(self, path: str) -> str:
        return self.delegate.url_for(path)

    def get(self, path: str, **query: Any) -> Response:
        self.record_request(Method.GET, path, **query)
        resp = self.delegate.get(path, **query)
        self.record_response(resp)
        return resp

    def patch(self, path: str, **fields: Any) -> Response:
        self.record_request(Method.PATCH, path, **fields)
        resp = self.delegate.patch(path, **fields)
        self.record_response(resp)
        return resp

    def post(self, path: str, **entity: Any) -> Response:
        self.record_request(Method.POST, path, **entity)
        resp = self.delegate.post(path, **entity)
        self.record_response(resp)
        return resp

    def delete(self, path: str, **query: Any) -> Response:
        self.record_request(Method.DELETE, path, **query)
        resp = self.delegate.delete(path, **query)
        self.record_response(resp)
        return resp

    def record_request(self, method: Method, path: str, **params: Any):
        self.log.info(
            f"{self.delegate_name}: → "
            f"{method.value} "
            f"{self.url_for(path)} "
            f"{LoggingWireProtocol.format_json(params)}"
        )

    def record_response(self, resp: Response):
        try:
            payload = LoggingWireProtocol.format_json(resp.json())
        except json.JSONDecodeError:
            payload = repr(resp.content)
        self.log.info(f"{self.delegate_name}: ← " f"{resp.code} " f"{payload}")

    @staticmethod
    def format_json(d: Optional[Dict[str, Any]]) -> str:
        if d is None:
            return ""
        return json.dumps(d, sort_keys=True)


class LoggingHTTPWireProtocol(LoggingWireProtocol):
    """Specialisation of ``HTTPWireProtocol`` that logs HTTP requests."""

    def __init__(self, *args, **kwargs):
        super().__init__(HTTPWireProtocol(*args, **kwargs))

    @property
    def session(self) -> requests.Session:
        transport = cast(HTTPWireProtocol, self.delegate)
        return transport.session


class NoopHTTPProtocol(LoggingHTTPWireProtocol):
    def patch(self, path: str, **fields: Any) -> Response:
        self.log.info(
            f"Suppressing PATCH {self.url_for(path)} "
            f"{LoggingWireProtocol.format_json(fields)}"
        )
        return HTTPResponse.empty()

    def post(self, path: str, **payload: Any) -> Response:
        self.log.info(
            f"Suppressing POST {self.url_for(path)} "
            f"{LoggingWireProtocol.format_json(payload)}"
        )
        return HTTPResponse.empty()

    def delete(self, path: str, **query: Any) -> Response:
        self.log.info(f"Suppressing DELETE {self.url_for(path)}")
        return HTTPResponse.empty()


def valid_url(url: str) -> bool:
    scheme, netloc, path, _, _, _ = urlparse(url)
    return all([scheme, netloc, path])
