from enum import Enum
from typing import (
    Any,
    Dict,
    Optional,
    Protocol,
    runtime_checkable,
)


class Response(Protocol):
    """Contains a server's response to a ``WireProtocol`` request."""

    @property
    def code(self) -> int:
        """Response HTTP status code."""
        ...

    @property
    def content(self) -> Optional[bytes]:
        """Response body represented as bytes."""
        ...

    @property
    def text(self) -> str:
        """Response body represented as a string."""
        ...

    def json(self, **kwargs) -> Any:
        """
        Response converted to JSON, if possible.

        :param kwargs:
            Same parameters accepted by ``json.loads``.

        :raises json.JSONDecodeError:
            If the content cannot be represented as JSON.
        """
        ...

    def raise_if_error(self):
        """Raises a ``dezoom.zoom.v2.ZoomError`` if the response is an error."""
        ...


class Method(Enum):
    GET = "GET"
    HEAD = "HEAD"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    PATCH = "PATCH"


@runtime_checkable
class WireProtocol(Protocol):
    def __init__(
        self, url: str, prefix: str = "/", headers: Dict[str, str] = None,
    ):
        ...

    def url_for(self, path: str) -> str:
        """Join ``path`` together with the base URL, producing an absolute URL."""
        ...

    def get(self, path: str, **query: Any) -> Response:
        """Request representation of a resource."""
        ...

    def patch(self, path: str, **fields: Any) -> Response:
        """Apply partial modifications to a resource."""
        ...

    def post(self, path: str, **entity: Any) -> Response:
        """
        Submit an entity to the specified resource, often causing
        a change in state or side effects on the server.
        """
        ...

    def delete(self, path: str, **query: Any) -> Response:
        """Delete specified resource."""
        ...
