from collections import defaultdict
from json import JSONDecodeError
from typing import (
    Dict,
    List,
    Optional,
    Type,
)

from pydantic import ValidationError

from dezoom.zoom import BaseModel
from dezoom.zoom.typing import Response


class ErrorField(BaseModel):
    field: str
    message: str


class ErrorModel(BaseModel):
    code: int
    message: str
    errors: Optional[List[ErrorField]]


class ZoomError(Exception):
    """All Zoom errors inherit from this base class."""

    code: int = -1
    errors: List[ErrorField] = []
    cause: Optional[Exception] = None

    def __init__(
        self,
        message: Optional[str] = None,
        errors: Optional[List[ErrorField]] = [],
        cause: Optional[Exception] = None,
    ):
        super().__init__(message)
        self.message = message
        if errors is not None:
            self.errors = errors
        self.cause = cause

    def to_model(self) -> ErrorModel:
        return ErrorModel(code=self.code, message=self.message, errors=self.errors)


class UnknownError(ZoomError):
    """Fallback if the error code is not recognised."""


class InvalidData(ZoomError):
    """Invalid or missing data."""

    code = 400


class InvalidCredentials(ZoomError):
    """Invalid or missing credentials."""

    code = 401


class NotFound(ZoomError):
    """
    The resource does not exist, e.g. requesting a resource using
    an invalid or non-existent user ID.
    """

    code = 404


class DataConflict(ZoomError):
    """
    Raised when trying to overwrite a resource, e.g. when creating
    a user with an email that already exists.
    """

    code = 409


class RateLimit(ZoomError):
    """Raised when the client hits an API rate limitation."""

    code = 429


class GroupNotFound(ZoomError):
    """A group with this group ID does not exist."""

    code = 4130


class GroupMemberNotFound(ZoomError):
    """A group member with this ID does not exist."""

    code = 4131


def from_http_code(
    error_code: int, default: Type[ZoomError] = UnknownError
) -> Type[ZoomError]:
    """
    Gets error from its `error_code`, falling back to `default` if
    the error code is not recognised.
    """
    return _errors.get(error_code, default)


def from_response(resp: Response, default: Type[ZoomError] = UnknownError) -> ZoomError:
    """Constructs error from the basis of an HTTP response."""
    try:
        raw_error = ErrorModel.parse_raw(resp.text)
        code = raw_error.code
        message = raw_error.message
        fields = raw_error.errors
        cause = None
    except (JSONDecodeError, ValidationError) as e:
        code, message, fields, cause = resp.code, str(e), None, e
    cls = from_http_code(code, default=default)
    return cls(message, errors=fields, cause=cause)


_errors: Dict[int, Type[ZoomError]] = defaultdict()
for item in list(locals().values()):
    if type(item) == type and issubclass(item, ZoomError) and hasattr(item, "code"):
        _errors[item.code] = item
