from typing import (
    Iterable,
    List,
    Optional,
)

from dezoom.zoom import (
    BaseModel,
    Resource,
)
from dezoom.zoom.v2.users import LicenseType


class Group(BaseModel):
    id: str
    name: str
    total_members: int


class GroupMember(BaseModel):
    id: str
    first_name: str
    last_name: str
    email: str
    type: LicenseType


class GroupMembersWrapper(BaseModel):
    members: List[GroupMember]
    page_count: Optional[int]
    page_number: Optional[int]
    page_size: int
    total_records: int
    next_page_token: Optional[str]


class Groups(Resource):
    def __iter__(self) -> Iterable[Group]:
        """List of groups."""

        class GroupsWrapper(BaseModel):
            groups: List[Group]

        resp = self.transport.get("/groups")
        wrapper = GroupsWrapper.parse_raw(resp.text)
        return iter(wrapper.groups)

    def get(self, id: str) -> Group:
        """Get a group under an account."""
        resp = self.transport.get(f"/groups/{id}")
        return Group.parse_raw(resp.text)

    def add_member(self, group_id: str, user_id: str):
        """Add user as member to group."""
        self.transport.post(f"/groups/{group_id}/members", members=[{"id": user_id}])

    def remove_member(self, group_id: str, user_id: str):
        """Remove a member from group."""
        self.transport.delete(f"/groups/{group_id}/members/{user_id}")

    def list_members(self, group_id: str) -> Iterable[GroupMember]:
        """List the members of a group."""

        def get_page(next_page_token=None) -> GroupMembersWrapper:
            params = {}
            if next_page_token:
                params["next_page_token"] = next_page_token
            resp = self.transport.get(f"/groups/{group_id}/members", **params)
            return GroupMembersWrapper.parse_raw(resp.text)

        def unwrap(page: GroupMembersWrapper) -> Iterable[GroupMember]:
            for member in page.members:
                yield member
            if page.next_page_token:
                next_page = get_page(next_page_token=page.next_page_token)
                yield from unwrap(next_page)

        first_page = get_page()
        yield from unwrap(first_page)
