from enum import Enum
from typing import Iterable, List, Literal, Optional

from dezoom.zoom import (
    BaseModel,
    Resource,
)


class LicenseType(Enum):
    BASIC = 1
    LICENSED = 2
    ON_PREM = 3

    def __repr__(self):
        return f"{self.__module__}.{type(self).__name__}.{self.name}"


class User(BaseModel):
    id: str
    first_name: str
    last_name: str
    email: str
    type: LicenseType
    group_ids: Optional[List[str]]


class UsersWrapper(BaseModel):
    users: List[User]
    page_count: Optional[int]
    page_number: Optional[int]
    page_size: int
    total_records: int
    next_page_token: Optional[str]


class Users(Resource):
    def get(self, id: str) -> User:
        """
        A Zoom account can have one or more users. Use this API to
        view information of a specific user on a Zoom account.

        :param id:
            User ID to look up.
        """
        resp = self.transport.get(f"/users/{id}")
        return User.parse_raw(resp.text)

    def list(self) -> Iterable[User]:
        """List of users."""

        def get_page(next_page_token=None) -> UsersWrapper:
            params = {"page_size": 100}
            if next_page_token:
                params["next_page_token"] = next_page_token
            resp = self.transport.get(f"/users", **params)
            return UsersWrapper.parse_raw(resp.text)

        def unwrap(page: UsersWrapper) -> Iterable[User]:
            for user in page.users:
                yield user
            if page.next_page_token:
                next_page = get_page(next_page_token=page.next_page_token)
                yield from unwrap(next_page)

        first_page = get_page()
        yield from unwrap(first_page)

    # TODO(andretol): consider an abstraction between the model
    # and the returned resource so that changes could be reflected
    # without an explicit API
    def set_license_type(self, user_id: str, type: LicenseType):
        """Set a Zoom user's license type."""
        self.transport.patch(f"/users/{user_id}", type=type.value)

    def delete(
        self, user_id: str, action: Literal["delete", "disassociate"] = "delete"
    ):
        """Delete user."""
        self.transport.delete(f"/users/{user_id}", action=action)
